/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Michael H. wrote this file.  As long as you retain this notice you can
 * do whatever you want with this stuff. If we meet some day, and you
 * think this stuff is worth it, you can buy me a beer in return.       Michael
 * ----------------------------------------------------------------------------
 */

/*
 * ----------------------------------------------------------------------------
 * Das Vorhaben:
 * Vier Servos sollen in einer bestimmten Reihenfolge jeweils zwei vorher fest-
 * legbare End-Positionen anfahren, wenn vom Fernbedienempfnger ein Signal mit
 * hohem oder niedrigem Tastverhltnis ankommt.
 * ----------------------------------------------------------------------------
 * Die generelle Idee:
 * Die vier Servos and PRT_SV sollen unabhngig angesteuert werden knnen.
 * Timer 0 luft alle 10ms ber und stt alle 2*10ms die Erzeugung der Servo-
 * Pulse mittels Timer2 an. Noch in Timer 0 wird der erste Pin fr die PWM der
 * Servos gesetzt.
 * Timer 2 luft im CTC (Clear Timer on Compare) Mode und schaltet bei berlauf
 * den letzten PWM-Pin aus und den nchsten an, bis alle vier Servos bedient
 * wurden. So werden *nacheinander* vier Pulse alle 20ms aus Compare Matches
 * erzeugt. Vorteil: keine Sortierung oder hhere Auflsung ntig. Nicht Prob-
 * lematischer Nachteil: 20ms Persiode werden nicht exakt eingehalten.
 * ----------------------------------------------------------------------------
 * Zustzlich kann in einem zweiten Betriebsmodus die Endpositionen fr jeden
 * Servo mittels eines Spannungsteiler-Potis am ADC Pin P_SP festgelegt und im
 * EEPROM gespeichert werden. Die Auswahl des einzustellenden Servos erfolgt
 * ber einen Hec-Encoding Switch an PIN_SVS.
 * ----------------------------------------------------------------------------
 * In Timer 0 ist auerdem ein Timeout realisiert, der die Puls-Erzeugung und
 * damit die Servos abschaltet, falls sich nicht innerhalb von TIMEOUT_RCI ein
 * Puls an RCI meldet. Ist die Fernbedienung also aus, werden die Servos nicht
 * an Position gehalten.
 * ----------------------------------------------------------------------------
 */



#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

#include "pindefs.h"
#include "makrodefs.h"
#include "servodefs.h"

#include "endpos.h"


/* Timeout for a Pulse from the remote conrol receiver to show up	*
 * time [s] = timeout * 10ms										*/
#define		TIMEOUT_RCI		100

/* ABSOLETE allowed minimum and maximum servo position in terms		*
 * of sv_pos														*
 * 46 = 602 us    -    180 = 2,39 ms								*/
#define		SV_POS_MIN		46
#define		SV_POS_MAX		186

/* global var for current servo position				*
 * stray last value due to "speciality" of my code ^^	*/
volatile unsigned char sv_pos[3+1] = {0,0,0,0};
volatile unsigned char sv_end_pos[3*2];
/* pulse length [us] = 12.80 * endpos + 12.375	*
 * endpos = pulse length [us] / 12.80 -  1.045	*/
/* 1000us - 77  *
 * 1500us - 116 *
 * 2000us - 155 */

/* current gear position flag set in the INT0 interrupt after 		*
 * remote control receiption. initial position is down				*/
volatile char sv_pos_gear=1;

/* global timeout flag for receivings */
volatile unsigned char timeout=TIMEOUT_RCI+1;



void setup_mcu(void)
{
	/* set data directions and pullups */
	DDR_LED |= (1<<P_LED0)|(1<<P_LED1);
	PRT_SAV |= (1<<P_SAV);
	PRT_SVS |= ((1<<P_SVS0)|(1<<P_SVS1)|(1<<P_SVS2)|(1<<P_SVS3));

	/* timer 0							*
	 * set up for compare match at 10ms	*/
	/* CS02 CS01 CS00				*
	*    0    0    0    stopped		*
	*    0    0    1    fosc		*
	*    0    1    0    fosc/8		*
	*    0    1    1    fosc/64		*
	*    1    0    0    fosc/256	*
	*    1    0    1    fosc/1024	*/
	OCR0A  = 195;
	TIMSK0 = (1<<OCIE0A);
	TCCR0A = (1<<WGM01);
	TCCR0B = (1<<CS02)|(1<<CS00);

	/* timer 2														*
	 * set up for compare match to get maximum resolution over		*
	 * 2.5ms pulse width out of 1 byte: (20e6)/1/256 * 2.5e-3 = 195	*
	 * so there will be 195 steps over more than the standardized	*
	 * servo pulse width length										*/
	/* CS22 CS21 CS20				*
	*    0    0    0    stopped		*
	*    0    0    1    fosc		*
	*    0    1    0    fosc/8		*
	*    0    1    1    fosc/32		*
	*    1    0    0    fosc/64		*
	*    1    0    1    fosc/128	*
	*    1    1    0    fosc/256	*
	*    1    1    1    fosc/1024	*/
	TIMSK2 = (1<<OCIE0A);
	/* CTC */
	TCCR2A = (1<<WGM21);
	/* no prescaler */
	/* done in code via macro tmr2_on() */
	// TCCR2B = (1<<CS22)|(1<<CS21);

	/* external interrupt INT0 */
	/* done in code via macro int1_on() */
	EICRA = (1<<ISC10);
	EIMSK = (1<<INT1);

	/* setup ADC													*
	 * left adjusted conversion result, because who needs 10 bits	*
	 * anyway.														*
	 * ADC chan set to P_SP. ref is AVcc. clock devider set to 64	*/
	DIDR0 = (1<<ADC1D);
	ADMUX = (1<<REFS0)|(1<<ADLAR)|(1<<MUX0);
	ADCSRA= (1<<ADEN)|(1<<ADPS1)|(1<<ADPS0);
}


ISR(TIMER0_COMPA_vect)
{	/* every 10ms */
	static unsigned char i=0, j=0;
	if(i%2)
	{	/* every 2*10ms 											*
		 * initiate pulse-generation - set the first pulse length 	*/
		OCR2A  = sv_pos[0];
		PRT_SV |= (1<<P_SV0);
		tmr2_on();
	}
	i++;
	
	/* alive led - toggle every 256/8*10ms=320ms */
	j += 8;
	if(!j)	led_gn_tgl();

	timeout++;
	if(timeout>TIMEOUT_RCI)
	{	/* nothing received from the remote control since 100*10ms,	*
		 * so turn off the servo signals							*/
		tmr0_off();
		led_gn_aus();
	}
	else if(!(TCCR0B&(1<<CS00)))	tmr0_on();
}

ISR(TIMER2_COMPA_vect)
{	/* pulse generation												 *
	 * after pulse length has been reached, the interrupt occurs, so *
	 * clear the last servo-portpin, set the new portpin 			 */
	static unsigned char sv_idx=0;

	/* clear servo-portpin, stop timer */
	PRT_SV &= ~(0x70>>P_SV0);
	
	tmr2_off();
	sv_idx++;

	/* set pulse length for next servo*/
	OCR2A  = sv_pos[sv_idx];
	/* set the next servo-portpin */
	PRT_SV |= (1<<((P_SV0+sv_idx)));
	tmr2_on();

	if(sv_idx==3)
	{	/* all servos served, so stop counter, reset index */
		sv_idx=0;
		tmr2_off();
	}
}


ISR(INT1_vect)
{	/* used for pulse length detection incoming from the remote		*
	 * control receiver												*/
	/* buffer for the timer count at the occurring inverted rising	*
	 * edge															*/
	static unsigned char tmr0_cnt;
	/* first timer count at the inverted falling edge, then the		*
	 * time difference between inverted rising and falling edge		*/
	unsigned char tmr0_tmp;

	if(!(PRT_RCI&(1<<P_RCI)))
	{	/* inverted rising edge -> save timer counter */
		tmr0_cnt = TCNT0;
		/* since there's obviously a remote control present, reset	*
		 * the timeout counter and re-enable the timer if needed	*/
		timeout_res();
	}
	else
	{	/* inverted falling edge -> do the check */
		tmr0_tmp = TCNT0;
		if(tmr0_tmp>tmr0_cnt)
		{	/* no timer overflow occured in the mean time from		*
			 * rising to falling edge, everything's fine			*/
			{tmr0_tmp = tmr0_tmp - tmr0_cnt; GPIOR0=tmr0_tmp;}
		}
		else	{tmr0_tmp = (OCR0A-tmr0_cnt) + tmr0_tmp +1; GPIOR0=tmr0_tmp;}

		if(tmr0_tmp<29)
		{	/* pulse length between 20 and 39 counter ticks.		*
			 * low pulse length -> gear position down (else: up)	*/
			sv_pos_gear=0;
		}
		else 	sv_pos_gear=1;
	}
}



int main (void)
{
	unsigned char sv_pos_gear_last = ~sv_pos_gear;
	unsigned char sv_mov_dly0, sv_mov_dly1;

	/* init mcu */
	setup_mcu();
	
	/* get values for servo end-positions and delay times from		*
	 * the eeprom 													*
	 * set the unknown starting servo position to a save state		*/
	for(unsigned char i=0;i<6;i++)	sv_end_pos[i] = eep_rd(i);
	sv_mov_dly0 = eep_rd(13);
	sv_mov_dly1 = eep_rd(14);

	/* if set switch is still on, wait, doing nothing */
	while(!(PIN_SET&(1<<P_SET)))	led_rt_an();
	led_rt_aus();
	/* initialization done, turn on servos by... */
	/* ...global interrupts enable */
	sei();
	sv_on();

	while(1)
	{
		if(!(PIN_SET&(1<<P_SET)))
		{	/* go to end-position adjustment mode */
			endpos_set_mode();
			/* update values just written to the eeprom while in 	*
			 * adjustment mode to the non global ram variables		*/
			sv_mov_dly0 = eep_rd(13);
			sv_mov_dly1 = eep_rd(14);
		}

		if(sv_pos_gear != sv_pos_gear_last)
		{	/* gear position needs to be changed */
			/* first check, if movement is neccessarry at all since	*
			 * it just adds unavoidable waiting time after an un-	*
			 * neccessary move										*/
			if(sv_pos_gear)
			{	/* get gear down */
				if( sv_pos[HECK_KLAPPE] != sv_end_pos[HECK_KLAPPE_AUF] ||
					sv_pos[BUG_KLAPPE]  != sv_end_pos[BUG_KLAPPE_AUF])
				{
					sv_pos[HECK_KLAPPE] = sv_end_pos[HECK_KLAPPE_AUF];
					sv_pos[BUG_KLAPPE]  = sv_end_pos[BUG_KLAPPE_AUF];
					if(wait_check(sv_mov_dly0))		{sv_pos_gear_last=~sv_pos_gear; continue;}
				}
				if(sv_pos[FAHRW] != sv_end_pos[FAHRW_AUF])
				{
					sv_pos[FAHRW] = sv_end_pos[FAHRW_AUF];
					if(wait_check(sv_mov_dly1))		{sv_pos_gear_last=~sv_pos_gear; continue;}
				}
				if(sv_pos[HECK_KLAPPE] != sv_end_pos[HECK_KLAPPE_ZU])
				{
					sv_pos[HECK_KLAPPE] = sv_end_pos[HECK_KLAPPE_ZU];
					if(wait_check(sv_mov_dly0))		{sv_pos_gear_last=~sv_pos_gear; continue;}
				}
			}
			else
			{	/* get gear up */
				if( sv_pos[HECK_KLAPPE] != sv_end_pos[HECK_KLAPPE_AUF] ||
					sv_pos[BUG_KLAPPE]  != sv_end_pos[BUG_KLAPPE_AUF])
				{
					sv_pos[HECK_KLAPPE] = sv_end_pos[HECK_KLAPPE_AUF];
					sv_pos[BUG_KLAPPE]  = sv_end_pos[BUG_KLAPPE_AUF];
					if(wait_check(sv_mov_dly0))		{sv_pos_gear_last=~sv_pos_gear; continue;}
				}
				if(sv_pos[FAHRW] != sv_end_pos[FAHRW_ZU])
				{
					sv_pos[FAHRW] = sv_end_pos[FAHRW_ZU];
					if(wait_check(sv_mov_dly1))		{sv_pos_gear_last=~sv_pos_gear; continue;}
				}
				if( sv_pos[HECK_KLAPPE] != sv_end_pos[HECK_KLAPPE_ZU] ||
					sv_pos[BUG_KLAPPE]  != sv_end_pos[BUG_KLAPPE_ZU])
				{
					sv_pos[HECK_KLAPPE] = sv_end_pos[HECK_KLAPPE_ZU];
					sv_pos[BUG_KLAPPE]  = sv_end_pos[BUG_KLAPPE_ZU];
					if(wait_check(sv_mov_dly0))		{sv_pos_gear_last=~sv_pos_gear; continue;}
				}
			}
			sv_pos_gear_last = sv_pos_gear;
		}
	}
	return 0;
}
